# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Gettings Things Gnome! - a personal organizer for the GNOME desktop
# Copyright (c) 2008-2009 - Lionel Dricot & Bertrand Rousseau
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# -----------------------------------------------------------------------------

#=== IMPORT ===================================================================

#system imports
import pygtk
pygtk.require('2.0')
import gobject
import glib
import gtk
import cairo
import pango
from math import pi, ceil
from GTG.tools.logger import Log
from GTG.tools.dates  import Date
from GTG.core.task    import Task

#=== MAIN CLASS ===============================================================

def get_contrast_color(color):
    col_yiq = ((color.red*299)+(color.green*587)+(color.blue*114)) / 1000.
    if col_yiq >= 32768:
        return gtk.gdk.color_parse("#2E3436")
    else:
        return gtk.gdk.color_parse("#D3D7CF")

class TagRenderer:

    XPAD = 4
    YPAD = 2

    def __init__(self, gc, cr, pc):
        self.gc = gc
        self.cr = cr
        self.layout = pango.Layout(pc)
        self.tag = None
        self.icon_size = 16
        self.tag_area  = gtk.gdk.Rectangle(0, 0, 0, 0)
        self.txt_area  = gtk.gdk.Rectangle(0, 0, 0, 0)

    def __rounded_rec(self, context, x, y, w, h, r = 10):
        "Draw a rounded rectangle"
        context.new_sub_path()
        context.arc(x+w-r, y+r  , r+1, -0.5*pi,      0)
        context.arc(x+w-r, y+h-r, r+1,       0, 0.5*pi)
        context.arc(x+r  , y+h-r, r+1,  0.5*pi, 1.0*pi)
        context.arc(x+r  , y+r  , r+1,  1.0*pi, 1.5*pi)
        context.close_path()

    # private methods

    def __update(self):
        self.__update_pango_layout()
        self.__update_txt_area()
        self.__update_tag_area()

    def __update_pango_layout(self):
        attr = pango.AttrList()
        size = pango.AttrSize(8 * pango.SCALE, 0, -1)
        attr.insert(size)
        self.layout.set_attributes(attr)

    def __update_txt_area(self):
        if self.tag is None:
            txt_width  = 0
            txt_height = 0
        else:
            # txt_height is tricky: it requires to evaluate correctly the font
            # descent, so we estimate it on a string that contains enough
            # glyphs with non-zero descent: "glyph"
            self.layout.set_text("glyph")
            txt_y      = 0
            txt_height = pango.DESCENT(self.layout.get_pixel_extents()[0])
            # and we restore
            self.layout.set_text(self.tag.get_name()[1:])
            txt_x     = -1 * pango.LBEARING(self.layout.get_pixel_extents()[0])
            txt_width = self.layout.get_pixel_extents()[0][2]
        self.txt_area = gtk.gdk.Rectangle(txt_x, txt_y, txt_width, txt_height)

    def __update_tag_area(self):
        if self.tag is None:
            tag_width  = 0
            tag_height = 0
        else:
            icon = self.tag.get_attribute("icon")
            if icon is not None:
                tag_width  = self.icon_size
                tag_height = self.icon_size
            else:
                tag_width  = 2*self.XPAD + self.txt_area[2]
                tag_height = 2*self.YPAD + self.txt_area[3]
        self.tag_area = gtk.gdk.Rectangle(0, 0, tag_width, tag_height)

    # configuration accessor

    def set_tag(self, tag):
        self.tag = tag
        self.__update()

    def get_tag(self):
        return self.tag

    # rendering-related methods

    def get_tag_area(self):
        return self.tag_area

    def render(self, window, widget, render_area):
        ctx = gtk.gdk.CairoContext(self.cr)
        # retrieve sizes and check sanity
        orig_x, orig_y = render_area[0], render_area[1]
        x, y, width, height = self.get_tag_area()
        # color, icon
        icon  = self.tag.get_attribute("icon")
        # draw icon
        if icon is not None:
            pixbuf = gtk.icon_theme_get_default().load_icon( \
                icon, self.icon_size, 0)
            icon_x = orig_x
            icon_y = orig_y + (height-self.icon_size)/2
            ctx.set_source_pixbuf(pixbuf, icon_x, icon_y)
            ctx.paint()
        # draw text label
        else:
            color = self.tag.get_attribute("color")
            if color is None:
                color = "#FFFFFF"
            color = gtk.gdk.color_parse(color)
            # draw background shape
            self.__rounded_rec(ctx, orig_x+0.5, orig_y+0.5, width-2, height-2, r=1.0)
            ctx.set_source_color(color)
            ctx.fill()
            # draw shape border
            ctx.set_source_rgba(0, 0, 0, 0.15)
            self.__rounded_rec(ctx, orig_x+0.5, orig_y+0.5, width-2, height-2, r=1.0)
            ctx.stroke()
            txt_x = orig_x+self.XPAD+self.txt_area.x
            txt_y = orig_y+int((height-2-self.txt_area.height)/2)
            txt_col = get_contrast_color(color)
            attr = self.layout.get_attributes()
            pa_fgcolor = pango.AttrForeground(txt_col.red, txt_col.green, txt_col.blue, 0, -1)
            attr.insert(pa_fgcolor)
            self.layout.set_attributes(attr)
            window.draw_layout(self.gc, txt_x, txt_y, self.layout, None, None)


class CellRendererTask(gtk.GenericCellRenderer):
    __gproperties__ = {
        'task': (gobject.TYPE_PYOBJECT,\
            "Task", "The cell task", gobject.PARAM_READWRITE),
    }

    DFT_HEIGHT      = 42
    DFT_DATE_OFFSET = 120
    ICON_SIZE       = 16

    # Private methods
    def __roundedrec(self, context, x, y, w, h, r = 10):
        "Draw a rounded rectangle"
        context.new_sub_path()
        context.arc(x+w-r, y+r  , r+1, -0.5*pi,      0)
        context.arc(x+w-r, y+h-r, r+1,       0, 0.5*pi)
        context.arc(x+r  , y+h-r, r+1,  0.5*pi, 1.0*pi)
        context.arc(x+r  , y+r  , r+1,  1.0*pi, 1.5*pi)
        context.close_path()
        return

    # Class methods
    def __init__(self): #pylint: disable-msg=W0231
        self.__gobject_init__()
        self.task = None

    def do_set_property(self, pspec, value):
        if pspec.name == "task":
            self.task = value
        else:
            setattr(self, pspec.name, value)

    def do_get_property(self, pspec):
        if pspec.name == "task":
            return self.task
        else:
            return getattr(self, pspec.name)

    def on_render(\
        self, window, widget, background_area, cell_area, expose_area, flags):
        cell_rect = gtk.gdk.Rectangle()
        cell_rect.x, cell_rect.y, cell_rect.width, cell_rect.height = \
            self.on_get_size(widget, cell_area)

        # Drawing context
        gc         = window.new_gc()
        cr         = window.cairo_create()
        pc         = widget.get_pango_context()
        gdkcontext = gtk.gdk.CairoContext(cr)
        gdkcontext.set_antialias(cairo.ANTIALIAS_DEFAULT)

        # Coordinates of the origin point
        x_align = self.get_property("xalign")
        y_align = self.get_property("yalign")
        orig_x  = cell_area.x
        orig_y  = cell_area.y

        # Draw the background
        rr_x      = orig_x + self.get_property("xpad")
        rr_y      = orig_y + self.get_property("ypad")
        if (rr_x*2) % 2 == 0:
            rr_x += 0.5
        if (rr_y*2) % 2 == 0:
            rr_y += 0.5
        rr_width  = cell_area.width - 2*self.get_property("xpad")
        rr_height = cell_area.height - 2*self.get_property("ypad")
        gdkcontext.set_source_rgba(0, 0, 0, 0.2)
        self.__roundedrec(cr, rr_x+1, rr_y+1, rr_width, rr_height, r = 2)
        gdkcontext.fill()
        my_color = gtk.gdk.color_parse("#FFFFFF")
        gdkcontext.set_source_color(my_color)
        self.__roundedrec(cr, rr_x, rr_y, rr_width, rr_height, r = 2)
        gdkcontext.fill()
        gdkcontext.set_line_width(1)
        my_color = gtk.gdk.color_parse("#BCBCBC")
        gdkcontext.set_source_color(my_color)
        self.__roundedrec(cr, rr_x, rr_y, rr_width, rr_height, r = 2)
        gdkcontext.stroke()

#        # Draw the check
#        if self.task.get_status() == Task.STA_ACTIVE:
#            pixbuf = gtk.icon_theme_get_default().load_icon( \
#                "gtg-task-undone", 24, 0)
#        else:
#            pixbuf = gtk.icon_theme_get_default().load_icon( \
#                "gtg-task-done", 24, 0)
#        icon_x = orig_x + self.get_property("xpad") + 8
#        icon_y = orig_y + (cell_area.height-24)/2
#        gdkcontext.set_source_pixbuf(pixbuf, icon_x, icon_y)
#        gdkcontext.paint()

        # Compute tag length (for title ellipsis)
        tagr = TagRenderer(gc, cr, pc)
        tag_lst = self.task.get_tags()
        tag_w = 0
        for i in xrange(len(tag_lst)):
            tag = tag_lst[i]
            if tag.get_attribute("color") is not None or \
               tag.get_attribute("icon") is not None:
                # tag text extents
                tagr.set_tag(tag)
                tmp_x, tmp_y, tmp_w, tmp_h = tagr.get_tag_area()
                tag_w += tmp_w + 2

        layout = pango.Layout(pc)
        font_desc = pango.FontDescription()
        font_desc.set_size(9 * pango.SCALE)
        layout.set_font_description(font_desc)
        layout.set_width( (self.DFT_DATE_OFFSET-16-8) * pango.SCALE)
        layout.set_text("glyph")
        txt_height = pango.DESCENT(layout.get_pixel_extents()[0])

        # Task start date, if any
        sdate = self.task.get_start_date()
        # icon
        pixbuf = gtk.icon_theme_get_default().load_icon( \
            "gtg-task-start-date", self.ICON_SIZE, 0)
        icon_x = orig_x + self.get_property("xpad") + rr_width - self.DFT_DATE_OFFSET + 5
        icon_y = orig_y + self.get_property("ypad") + 2 + int((rr_height/2-3-self.ICON_SIZE))
        gdkcontext.set_source_pixbuf(pixbuf, icon_x, icon_y)
        gdkcontext.paint()
        # text
        if sdate != Date.no_date():
            text = sdate.to_readable_string().title()
        else:
            text = "-"
        layout.set_text(text)
        txt_x = icon_x + 16 + 8
        txt_y = orig_y + self.get_property("ypad") + 2 + int((rr_height/2-3-txt_height)/2)
        attr = pango.AttrList()
        if Date.today() > sdate:
            pa_fgcolor = pango.AttrForeground(61440, 20480, 20480, 0, -1)
        else:
            pa_fgcolor = pango.AttrForeground(34815, 33790, 34047, 0, -1)
        attr.insert(pa_fgcolor)
        layout.set_attributes(attr)
        window.draw_layout(gc, txt_x, txt_y, layout, None, None)

        # Task due date, if any
        ddate = self.task.get_due_date()
        # icon
        pixbuf = gtk.icon_theme_get_default().load_icon( \
            "gtg-task-due-date", self.ICON_SIZE, 0)
        icon_x = orig_x + self.get_property("xpad") + rr_width - self.DFT_DATE_OFFSET + 5
        icon_y = orig_y + self.get_property("ypad") + rr_height/2 + 1 + int((rr_height/2-3-self.ICON_SIZE))
        gdkcontext.set_source_pixbuf(pixbuf, icon_x, icon_y)
        gdkcontext.paint()
        # text
        if ddate != Date.no_date():
            text = ddate.to_readable_string().title()
        else:
            text = "-"
        layout.set_text(text)
        txt_x = icon_x + 16 + 8
        txt_y = orig_y + self.get_property("ypad") + rr_height/2 + 1 + int((rr_height/2-3-txt_height)/2)
        attr = pango.AttrList()
        if Date.today() > ddate:
            pa_fgcolor = pango.AttrForeground(61440, 20480, 20480, 0, -1)
        else:
            pa_fgcolor = pango.AttrForeground(34815, 33790, 34047, 0, -1)
        attr.insert(pa_fgcolor)
        layout.set_attributes(attr)
        window.draw_layout(gc, txt_x, txt_y, layout, None, None)

        # Task title
        txt_x = orig_x + self.get_property("xpad") + 5
        txt_y = orig_y + self.get_property("ypad") + 4
        my_color = gtk.gdk.color_parse("#000000")
        gdkcontext.set_source_color(my_color)
        layout = pango.Layout(pc)
        layout.set_width( (rr_width - self.DFT_DATE_OFFSET - tag_w - 10 - 10) * pango.SCALE)
        layout.set_ellipsize(pango.ELLIPSIZE_END)
        font_desc = pango.FontDescription()
        font_desc.set_weight("bold")
        font_desc.set_size(10 * pango.SCALE)
        layout.set_font_description(font_desc)
        layout.set_text(self.task.get_title())
        window.draw_layout(gc, txt_x, txt_y, layout, None, None)
        title_pxe = layout.get_pixel_extents()[0]
        title_width = title_pxe[2]
        title_height = title_pxe[3]

        # Tag label
        tag_lst = self.task.get_tags()
        tag_x = orig_x + self.get_property("xpad") + rr_width - self.DFT_DATE_OFFSET - tag_w - 10
        tag_y = orig_y + self.get_property("ypad") + 4
        for i in xrange(len(tag_lst)):
            tag = tag_lst[i]
            if tag.get_attribute("color") is not None or \
               tag.get_attribute("icon") is not None:
                # tag text extents
                tagr.set_tag(tag)
                tmp_x, tmp_y, tmp_w, tmp_h = tagr.get_tag_area()
                render_area = gtk.gdk.Rectangle(tag_x, tag_y, tmp_w, tmp_h)
                tagr.render(window, widget, render_area)
                tag_x += tmp_w + 2

        # Task excerpt
        txt_x = orig_x + self.get_property("xpad") + 5
        txt_y = orig_y + self.get_property("ypad") + 4 + title_height + 5 + int((rr_height - 4 - title_height - 5 - txt_height)/2)
        layout = pango.Layout(pc)
        layout.set_width( (rr_width - self.DFT_DATE_OFFSET - 10 - 10) * pango.SCALE)
        layout.set_ellipsize(pango.ELLIPSIZE_END)
        attr = pango.AttrList()
        pa_fgcolor = pango.AttrForeground(34815, 33790, 34047, 0, -1)
        attr.insert(pa_fgcolor)
        size = pango.AttrSize(8 * pango.SCALE, 0, -1)
        attr.insert(size)
        layout.set_attributes(attr)
        text = self.task.get_excerpt(strip_tags=True, strip_subtasks=True)
        layout.set_text(text.replace('\n', ' '))
        excerpt_width = layout.get_pixel_extents()[0][2]
        excerpt_height = layout.get_pixel_extents()[0][3]
        window.draw_layout(gc, txt_x, txt_y, layout, None, None)

    def on_get_size(self, widget, cell_area=None): #pylint: disable-msg=W0613
        width = 2 * self.get_property("xpad") + 0
        height = 2 * self.get_property("ypad") + self.DFT_HEIGHT
        return (0, 0, width, height)

gobject.type_register(CellRendererTask)
